import subprocess
import logging


def run_cmd(cmd):
    try:
        result = subprocess.run(
            cmd,
            shell=True,
            encoding="ascii",
            check=True,
        )
        return result.returncode == 0

    except:
        logging.error("Could not run command %s", cmd, exc_info=True)
        return False


def startup_postprocessor(enable):
    if enable:
        cmd = "sudo systemctl enable thermal-postprocess && sudo systemctl restart thermal-postprocess"
    else:
        # disable but start once so that it can finish any stale files that may exist
        cmd = "sudo systemctl disable thermal-postprocess && sudo systemctl restart thermal-postprocess"
    return run_cmd(cmd)


def stop_network_classifier():
    cmd = "sudo systemctl stop thermal-classifier"
    return run_cmd(cmd)


def startup_network_classifier(enable):
    if enable:
        cmd = "sudo systemctl enable thermal-classifier && sudo systemctl start thermal-classifier"
    else:
        cmd = "sudo systemctl disable thermal-classifier && sudo systemctl stop thermal-classifier"
    return run_cmd(cmd)


def is_service_running(service_name):
    result = subprocess.run(["systemctl", "is-active", "--quiet", service_name])
    return result.returncode == 0



import gc
import json
import logging
import os.path
import time
import math
import numpy as np

import cv2
from classify.trackprediction import Predictions
from track.clip import Clip
from track.cliptrackextractor import ClipTrackExtractor, is_affected_by_ffc
from ml_tools import tools
from track.irtrackextractor import IRTrackExtractor
from ml_tools.previewer import Previewer
from ml_tools.interpreter import get_interpreter
from track.trackextractor import extract_file
from classify.thumbnail import get_thumbnail_info, best_trackless_thumb
from pathlib import Path


class ClipClassifier:
    """Classifies tracks within CPTV files."""

    # skips every nth frame.  Speeds things up a little, but reduces prediction quality.
    FRAME_SKIP = 1

    def __init__(
        self, config, model=None, keep_original_predictions=False, tracking_events=False
    ):
        """Create an instance of a clip classifier"""
        self.keep_original_predictions = keep_original_predictions
        self.config = config
        # super(ClipClassifier, self).__init__(config, tracking_config)
        self.model = model
        if self.keep_original_predictions:
            self.model.id = f"post-{self.model.id}"
            self.model.name = f"post-{self.model.name}"

        # prediction record for each track

        self.previewer = Previewer.create_if_required(config, config.classify.preview)

        self.models = {}
        self.tracking_events = tracking_events
        self._is_recording = False

    def set_is_recording(self, is_recording):
        self._is_recording = is_recording

    def load_models(self):
        for model in self.config.classify.models:
            logging.info("Loading %s", model)
            classifier = self.get_classifier(model)

    def get_classifier(self, model):
        """
        Returns a classifier object, which is created on demand.
        This means if the ClipClassifier is copied to a new process a new Classifier instance will be created.
        """
        if model.id in self.models:
            return self.models[model.id]
        load_start = time.time()
        logging.info("classifier loading %s", model.model_file)
        classifier = get_interpreter(model, model.run_over_network)
        logging.info("classifier loaded (%s)", time.time() - load_start)
        self.models[model.id] = classifier
        return classifier

    def get_meta_data(self, filename):
        """Reads meta-data for a given cptv file."""
        source_meta_filename = os.path.splitext(filename)[0] + ".txt"
        if os.path.exists(source_meta_filename):
            meta_data = tools.load_clip_metadata(source_meta_filename)

            tags = set()
            for record in meta_data["Tags"]:
                # skip automatic tags
                if record.get("automatic", False):
                    continue
                else:
                    tags.add(record["animal"])

            tags = list(tags)

            if len(tags) == 0:
                tag = "no tag"
            elif len(tags) == 1:
                tag = tags[0] if tags[0] else "none"
            else:
                tag = "multi"
            meta_data["primary_tag"] = tag
            return meta_data
        else:
            return None

    def process(
        self,
        source,
        cache=None,
        reuse_frames=None,
        track=False,
        calculate_thumbnails=False,
    ):
        # IF passed a dir extract all cptv files, if a cptv just extract this cptv file
        if not os.path.exists(source):
            logging.error("Could not find file or directory %s", source)
            return
        if os.path.isfile(source):
            self.process_file(
                source,
                cache=cache,
                reuse_frames=reuse_frames,
                track=track,
                calculate_thumbnails=calculate_thumbnails,
            )
            return
        for folder_path, _, files in os.walk(source):
            for name in files:
                if os.path.splitext(name)[1] in [".mp4", ".cptv", ".avi"]:
                    full_path = os.path.join(folder_path, name)
                    self.process_file(
                        full_path,
                        cache=cache,
                        reuse_frames=reuse_frames,
                        track=track,
                        calculate_thumbnails=calculate_thumbnails,
                    )

    def process_file(
        self,
        filename,
        cache=None,
        reuse_frames=None,
        track=False,
        calculate_thumbnails=False,
    ):
        """
        Process a file extracting tracks and identifying them.
        :param filename: filename to process
        :param enable_preview: if true an MPEG preview file is created.
        """
        _, ext = os.path.splitext(filename)
        cache_to_disk = (
            cache if cache is not None else self.config.classify.cache_to_disk
        )

        if track:
            logging.info("Doing tracking")
            clip, track_extractor = extract_file(
                filename, self.config, cache_to_disk, to_stdout=False
            )
        elif ext == ".cptv":
            track_extractor = ClipTrackExtractor(
                self.config.tracking,
                self.config.use_opt_flow,
                cache_to_disk,
                do_tracking=track,
                calculate_filtered=True,
                verbose=self.config.verbose,
                calculate_thumbnail_info=calculate_thumbnails,
            )
            logging.info("Using clip extractor")

        elif ext in [".avi", ".mp4"]:
            track_extractor = IRTrackExtractor(self.config.tracking, cache_to_disk)
            logging.info("Using ir extractor")
        else:
            logging.error("Unknown extention %s", ext)
            return False

        base_filename = os.path.splitext(os.path.basename(filename))[0]
        meta_file = os.path.join(os.path.dirname(filename), base_filename + ".txt")
        if not os.path.exists(filename):
            logging.error("File %s not found.", filename)
            return False
        if not os.path.exists(meta_file):
            logging.error("File %s not found.", meta_file)
            return False
        meta_data = tools.load_clip_metadata(meta_file)

        logging.info("Processing file '{}'".format(filename))

        if not track:
            clip = Clip(track_extractor.config, filename)
            clip.load_metadata(
                meta_data,
                self.config.build.tag_precedence,
            )
            track_extractor.parse_clip(clip)

        predictions_per_model = {}
        if self.model:
            prediction = self.classify_clip(
                clip,
                self.model,
                meta_data,
                reuse_frames=reuse_frames,
            )
            predictions_per_model[self.model.id] = prediction
        else:
            for model in self.config.classify.models:
                prediction = self.classify_clip(
                    clip,
                    model,
                    meta_data,
                    reuse_frames=reuse_frames,
                )
                predictions_per_model[model.id] = prediction
        destination_folder = os.path.dirname(filename)
        dirname = destination_folder

        if self.previewer:
            mpeg_filename = os.path.join(dirname, base_filename + "-classify.mp4")

            logging.info("Exporting preview to '{}'".format(mpeg_filename))

            self.previewer.export_clip_preview(
                mpeg_filename, clip, predictions_per_model
            )
        models = [self.model] if self.model else self.config.classify.models
        meta_data = self.save_metadata(
            meta_data,
            meta_file,
            clip,
            predictions_per_model,
            models,
            calculate_thumbnails=calculate_thumbnails,
        )
        if cache_to_disk:
            clip.frame_buffer.remove_cache()
        return meta_data

    def classify_clip(self, clip, model, meta_data, reuse_frames=None):
        start = time.time()
        classifier = self.get_classifier(model)
        predictions = Predictions(classifier.labels, model)
        predictions.model_load_time = time.time() - start

        for i, track in enumerate(clip.tracks):
            segment_frames = None
            if reuse_frames:
                tracks = meta_data.get("tracks")
                meta_track = next(
                    (x for x in tracks if x["id"] == track.get_id()), None
                )
                if meta_track is not None:
                    prediction_tag = next(
                        (
                            x
                            for x in meta_track["tags"]
                            if x.get("data", {}).get("name") == model.name
                        ),
                        None,
                    )
                    if prediction_tag is not None:
                        if "prediction_frames" in prediction_tag["data"]:
                            logging.info("Reusing previous prediction frames %s", model)
                            segment_frames = prediction_tag["data"]["prediction_frames"]
                            segment_frames = np.uint16(segment_frames)

            prediction = classifier.classify_track(
                clip, track, segment_frames=segment_frames, min_segments=1
            )
            if prediction is not None:
                predictions.prediction_per_track[track.get_id()] = prediction
                description = prediction.description()
                logging.info(
                    "{} - [{}/{}] prediction: {}".format(
                        track.get_id(), i + 1, len(clip.tracks), description
                    )
                )
        if self.config.verbose:
            ms_per_frame = (
                (time.time() - start) * 1000 / max(1, len(clip.frame_buffer.frames))
            )
            logging.info("Took {:.1f}ms per frame".format(ms_per_frame))
        if classifier.TYPE == "Keras":
            tools.clear_session()
        del classifier
        gc.collect()

        return predictions

    def save_metadata(
        self,
        meta_data,
        meta_filename,
        clip,
        predictions_per_model,
        models,
        calculate_thumbnails=False,
    ):
        tracks = meta_data.get("tracks")
        for track in clip.tracks:
            meta_track = next((x for x in tracks if x["id"] == track.get_id()), None)
            if meta_track is None:
                logging.error(
                    "Got prediction for track which doesn't exist in metadata"
                )
                continue
            prediction_info = []
            for model_id, predictions in predictions_per_model.items():
                prediction = predictions.prediction_for(track.get_id())
                if prediction is None:
                    continue
                prediction_meta = prediction.get_metadata()
                prediction_meta["model_id"] = model_id
                if self.keep_original_predictions:
                    prediction_meta["reprocessed"] = True
                prediction_info.append(prediction_meta)
            if self.keep_original_predictions:
                existing_predictions = meta_track.get("predictions", [])
                if existing_predictions is None:
                    existing_predictions = []
                prediction_info.extend(existing_predictions)
            meta_track["predictions"] = prediction_info

            if calculate_thumbnails:
                best_thumb, best_score = get_thumbnail_info(clip, track)
                if best_thumb is None:
                    meta_track["thumbnail"] = None
                else:
                    thumbnail_info = {
                        "region": best_thumb.region,
                        "contours": best_thumb.contours,
                        "median_diff": best_thumb.median_diff,
                        "score": round(best_score),
                    }
                    meta_track["thumbnail"] = thumbnail_info
        if calculate_thumbnails and len(clip.tracks) == 0:
            # if no tracks choose a clip thumb
            region = best_trackless_thumb(clip)
            meta_data["thumbnail_region"] = region

        model_dictionaries = {}

        for existing_model in meta_data.get("models", []):
            model_dictionaries[existing_model["id"]] = existing_model

        for model in models:
            if model.id in model_dictionaries:
                model_dic = model_dictionaries[model.id]
            else:
                model_dic = model.as_dict()
            model_predictions = predictions_per_model[model.id]
            model_dic["classify_time"] = float(
                round(
                    model_predictions.classify_time + model_predictions.model_load_time,
                    1,
                )
            )
            model_dictionaries[model.id] = model_dic
        meta_data["models"] = list(model_dictionaries.values())
        if self.config.classify.meta_to_stdout:
            logging.info("Printing json meta data")

            print(json.dumps(meta_data, cls=tools.CustomJSONEncoder))
        else:
            logging.info("saving meta data %s", meta_filename)
            with open(meta_filename, "w") as f:
                json.dump(meta_data, f, indent=4, cls=tools.CustomJSONEncoder)
        return meta_data

    def post_process_file(self, filename, service):
        from cptv_rs_python_bindings import CptvReader
        from piclassifier.motiondetector import RunningMean, SlidingWindow
        from piclassifier.cptvmotiondetector import CPTVMotionDetector
        from ml_tools.frame import Frame
        from ml_tools.preprocess import preprocess_frame, preprocess_movement
        from datetime import datetime
        import dbus

        filename = Path(filename)
        meta_file = filename.with_suffix(".txt")
        if not filename.exists():
            logging.error("File %s not found.", filename)
            return False
        if not meta_file.exists():
            logging.error("File %s not found.", meta_file)
            return False
        meta_data = tools.load_clip_metadata(meta_file)
        filename = Path(filename)
        meta_file = filename.with_suffix(".txt")
        if not filename.exists():
            logging.error("File %s not found.", filename)
            return False
        if not meta_file.exists():
            logging.error("File %s not found.", meta_file)
            return False
        meta_data = tools.load_clip_metadata(meta_file)

        rec_end = datetime.fromisoformat(meta_data["end_time"])

        # get segments here, or frames
        # only extra data for segments
        track_extractor = ClipTrackExtractor(
            self.config.tracking,
            self.config.use_opt_flow,
            calculate_filtered=True,
            verbose=self.config.verbose,
        )

        clip = Clip(track_extractor.config, filename)
        clip.load_metadata(
            meta_data,
            self.config.build.tag_precedence,
        )
        track_extractor.init_clip(clip)

        logging.info("Just running on first model")
        start = time.time()
        model = self.config.classify.models[0]
        classifier = self.get_classifier(model)
        predictions = Predictions(classifier.labels, model)
        predictions.model_load_time = time.time() - start

        track_samples = {}
        track_data = {}

        for track in clip.tracks:
            pred_frames = classifier.frames_for_prediction(clip, track)

            track_data[track.get_id()] = {
                "pred_frames": pred_frames,
                "limits": None,
                "frames": {},
                "track": track,
            }

            for seg in pred_frames:

                for r in seg.regions:
                    frame_data = track_samples.setdefault(r.frame_number, {})
                    frame_data[track.get_id()] = r
                    # frame_samples.append(r)
        reader = CptvReader(str(clip.source_file))
        current_frame_num = 0
        running_mean = None
        thermal_window = SlidingWindow(CPTVMotionDetector.MEAN_FRAMES, "O")

        if classifier.params.thermal_diff_norm:
            logging.error("Thermal min diff is not implemented so will not be used")

        while True:
            frame = reader.next_frame()

            if frame is None:
                break
            if frame.background_frame:
                continue

            if current_frame_num in track_samples:
                thermal_median = np.median(frame.pix)
                for track_id, region in track_samples[current_frame_num].items():
                    # region = track_samples[current_frame_num]
                    thermal = region.subimage(frame.pix).astype(np.float32)
                    background = region.subimage(
                        track_extractor.background_alg.background
                    )
                    filtered = thermal - background
                    thermal -= thermal_median
                    f = Frame(thermal, filtered, current_frame_num, region=region)
                    track_data[track_id]["frames"][region.frame_number] = f
                    if classifier.params.diff_norm:
                        f_min = np.min(filtered)
                        f_max = np.max(filtered)
                        existing_limits = track_data[track_id]["limits"]

                        if existing_limits is None:
                            track_data[track_id]["limits"] = [f_min, f_max]
                        else:
                            if f_min < existing_limits[0]:
                                existing_limits[0] = f_min
                            if f_max > existing_limits[1]:
                                existing_limits[1] = f_max
                            track_data[track_id]["limits"] = existing_limits
            # track_extractor.process_frame(clip, frame)
            is_ffc = is_affected_by_ffc(frame)
            oldest_thermal = thermal_window.oldest
            thermal_window.add(frame, is_ffc)

            if running_mean is None:
                running_mean = RunningMean([frame.pix], CPTVMotionDetector.MEAN_FRAMES)
            else:
                running_mean.add(frame.pix, oldest_thermal.pix)
            if not is_ffc:
                track_extractor.background_alg.process_frame(running_mean.mean())
            current_frame_num += 1
        i = 0
        for track_id, data in track_data.items():
            i += 1
            pred_frames = data["pred_frames"]
            pred_frame_numbers = []
            preprocessed = []
            masses = []
            for segment in pred_frames:
                segment_frames = []
                for frame_i in segment.frame_indices:
                    f = data["frames"][frame_i]
                    if not f.preprocessed:
                        f = preprocess_frame(
                            f,
                            (
                                classifier.params.frame_size,
                                classifier.params.frame_size,
                            ),
                            region,
                            clip.background,
                            clip.crop_rectangle,
                            calculate_filtered=False,
                            filtered_norm_limits=data["limits"],
                            cropped=True,
                            sub_median=False,
                        )
                        data["frames"][frame_i] = f
                    # probably no need to copy
                    segment_frames.append(f)
                frames = preprocess_movement(
                    segment_frames,
                    classifier.params.square_width,
                    classifier.params.frame_size,
                    classifier.params.channels,
                    classifier.preprocess_fn,
                    sample=f"{clip.get_id()}-{track_id}",
                )
                preprocessed.append(frames)
                masses.append(segment.mass)
                pred_frame_numbers.append(segment.frame_indices)
            if len(preprocessed) == 0:
                logging.info("No prediction made for track %s", track_id)
                continue
                # dont think this should happen
            preprocessed = np.array(preprocessed)

            # what to do if recording
            if self._is_recording:
                while self._is_recording:
                    logging.info("Waiting for current recording to finish")
                    time.sleep(10)
                # sleep here until not recording

            preds = []
            chunk_size = 5
            chunks = int(math.ceil(len(preprocessed) / chunk_size))
            # if is a very long track should break this up into samller chunks
            for chunk in range(chunks):
                preprocessed_chunk = preprocessed[
                    chunk * chunk_size : chunk * chunk_size + chunk_size
                ]
                logging.info(
                    "Predicting chunk %s (%s #) of %s %s:%s total preprocessed %s",
                    chunk,
                    len(preprocessed_chunk),
                    chunks,
                    chunk * chunk_size,
                    chunk * chunk_size + chunk_size,
                    len(preprocessed),
                )
                pred = classifier.predict(preprocessed_chunk)
                preds.extend(pred)
            track_prediction = classifier.track_prediction_from_raw(
                track_id, pred_frame_numbers, preds, masses
            )
            predictions.prediction_per_track[track_id] = track_prediction

            description = track_prediction.description()
            logging.info(
                "{} - [{}/{}] prediction: {}".format(
                    track_id, i, len(clip.tracks), description
                )
            )

            if (
                self.tracking_events
                and len(track_prediction.predictions) > 0
                # incase wasn't false positive during active tracking would want it reported now as FP
                # and track_prediction.predicted_tag() != "false-positive"
            ):
                dbus_preds = track_prediction.class_best_score.copy()
                dbus_preds = np.uint8(np.round(dbus_preds * 100))
                best = np.argmax(predictions)
                dbus_preds = dbus_preds.tolist()

                service.TrackReprocessed(
                    meta_data.get("id", 0),
                    track_id,
                    dbus_preds,
                    track_prediction.predicted_tag(),
                    int(round(100 * track_prediction.max_score)),
                    region.to_ltrb(),
                    region.frame_number,
                    region.mass,
                    region.blank,
                    True,
                    data["track"].bounds_history[-1].frame_number,
                    model.id,
                    rec_end.timestamp(),
                )

        models = [model]
        predictions_per_model = {model.id: predictions}

        meta_data = self.save_metadata(
            meta_data,
            meta_file,
            clip,
            predictions_per_model,
            models,
            calculate_thumbnails=False,
        )



import time
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
from config.thermalconfig import ThermalConfig
from pathlib import Path
import multiprocessing
from classify.clipclassifier import ClipClassifier
from config.config import Config
import logging
from ml_tools.logs import init_logging
import time
import sys
import shutil
from piclassifier import service
from functools import partial
import threading
import dbus
from gi.repository import GLib
from piclassifier.utils import startup_network_classifier, is_service_running


class DirWatcher(FileSystemEventHandler):
    def __init__(self, process_queue):
        self.process_queue = process_queue

    def on_created(self, event):
        if not event.is_directory:
            event_file = Path(event.src_path)
            if event_file.suffix == ".cptv":
                if event_file.with_suffix(".txt").exists():
                    self.process_queue.put(event_file)
            elif event_file.suffix == ".txt":
                if event_file.with_suffix(".cptv").exists():
                    self.process_queue.put(event_file.with_suffix(".cptv"))


def get_model(thermal_config, config):
    if not thermal_config.motion.run_classifier:
        logging.info("Classifier isn't configured to run in config")
        return None

    network_model = [
        model for model in config.classify.models if model.run_over_network
    ]
    if len(network_model) == 0:
        logging.info("No network model configured in classifier.yaml")
        return None
    if len(network_model) > 1:
        logging.info("Got multiple network models using first")
    return network_model[0]


def rec_callback(dt, is_recording, set_function):
    if is_recording:
        logging.info("Recording started pausing processing")
    else:
        logging.info("Recording ended resuming processing")
    set_function(is_recording)


def service_started():
    logging.info("Service started")
    # if we receive this means we need to reconnect, signals still work
    # but calling methods wont


def main():
    init_logging()
    config = Config.load_from_file()
    thermal_config = ThermalConfig.load_from_file()
    output_dir = Path(thermal_config.recorder.output_dir)
    reprocess_dir = output_dir / "postprocess"
    reprocess_dir.mkdir(exist_ok=True)
    network_model = get_model(thermal_config, config)
    if network_model is None:
        logging.info("No network model exiting")
        sys.exit(0)

    process_queue = multiprocessing.Queue()
    dir_watcher = DirWatcher(process_queue)
    observer = Observer()
    reprocess_files = list(reprocess_dir.glob("*.cptv"))
    for cptv_f in reprocess_files:
        logging.info("Adding existing %s", cptv_f)
        process_queue.put(cptv_f)

    postprocess = thermal_config.motion.postprocess
    pending_exit = False
    if not postprocess:
        if len(reprocess_files) == 0:
            logging.info(
                "No files to post process and config is not set to post process, exiting"
            )
            return
        logging.info(
            "Postprocessing stale files then exiting as config is not set to postprocess"
        )
        pending_exit = True
    else:
        logging.info("Watching %s", reprocess_dir)
        observer.schedule(dir_watcher, reprocess_dir, recursive=False)
        observer.start()
        config.validate()
    clip_classifier = ClipClassifier(
        config,
        network_model,
        keep_original_predictions=True,
        tracking_events=thermal_config.motion.postprocess_events,
    )

    callback_fn = partial(rec_callback, set_function=clip_classifier.set_is_recording)
    bus = None
    dbus_object = None
    need_dbus = thermal_config.motion.postprocess_events

    if need_dbus:
        max_attempts = 3
        attempt = 1
        while bus is None:
            try:
                dbus_object, bus, thread = connect_to_dbus(callback_fn)
            except Exception as ex:
                logging.info(
                    "Couldn't connecto dbus waiting 20 seconds and trying again",
                    exc_info=True,
                )
                if attempt >= max_attempts:
                    raise ex
                time.sleep(20)
            attempt += 1
    try:
        while True:
            try:
                new_file = process_queue.get(block=False)
            except:
                if pending_exit:
                    logging.info("Finished processing exit")
                    break
                time.sleep(5)
                continue
            new_file = Path(new_file)
            if not new_file.exists():
                continue
            # reprocess file

            if not is_service_running("thermal-classifier"):
                logging.info("Network classifier is not running starting it up")
                success = startup_network_classifier(True)
                if not success:
                    raise Exception("Could not start up netowrk classifier")
                # give it some time to start up
                time.sleep(5)
            try:
                if clip_classifier._is_recording:
                    while clip_classifier._is_recording:
                        logging.info(
                            "Waiting for current recording to finish before processing"
                        )
                        time.sleep(10)
                if need_dbus:
                    # ensures dbus service is always, can be invalidated if thermal-recorder restarts
                    try:
                        dbus_object = bus.get_object(
                            service.DBUS_NAME, service.DBUS_PATH
                        )
                    except:
                        logging.error(
                            "Could not connect to dbus service %s",
                            service.DBUS_NAME,
                            exc_info=True,
                        )
                        break
                clip_classifier.post_process_file(new_file, dbus_object)

            except:
                logging.error("Error reprocessing %s", new_file, exc_info=True)

            if new_file.exists():
                try:
                    meta_f = new_file.with_suffix(".txt")
                    shutil.move(new_file, output_dir / new_file.name)
                    shutil.move(meta_f, output_dir / meta_f.name)
                except:
                    logging.error(
                        "Error Moving %s to processing", new_file, exc_info=True
                    )

            logging.info("Finished processing %s", meta_f)

    except KeyboardInterrupt:
        observer.stop()
    loop.quit()
    if observer.is_alive:
        observer.stop()


def connect_to_dbus(rec_callback):
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    bus = dbus.SystemBus()
    dbus_object = bus.get_object(service.DBUS_NAME, service.DBUS_PATH)
    loop = GLib.MainLoop()

    dbus_thread = threading.Thread(target=dbus_events, args=(loop, bus, rec_callback))
    dbus_thread.start()
    return dbus_object, bus, dbus_thread


def dbus_events(loop, dbus_object, callback_fn):
    dbus_object.add_signal_receiver(
        callback_fn,
        dbus_interface=service.DBUS_NAME,
        signal_name="Recording",
    )
    dbus_object.add_signal_receiver(
        service_started,
        dbus_interface=service.DBUS_NAME,
        signal_name="ServiceStarted",
    )
    loop.run()



#!/usr/bin/env python3

import sys
from gi.repository import GLib

import dbus
import dbus.mainloop.glib
import time
import threading
import cv2
import numpy as np
import logging
from datetime import datetime
from pathlib import Path
from ml_tools.logs import init_logging


DBUS_NAME = "org.cacophony.thermalrecorder"
DBUS_PATH = "/org/cacophony/thermalrecorder"


model_labels = []
active_tracks = {}


def tracking(
    clip_id,
    track_id,
    prediction,
    what,
    confidence,
    region,
    frame,
    mass,
    blank,
    tracking,
    last_prediction_frame,
    model_id,
):
    logging.info(
        "Received tracking event for clip %s track %s prediction of %s with %s%% confidence still tracking ? %s at region %s",
        clip_id,
        track_id,
        what,
        confidence,
        tracking,
        region,
    )


def normalize(thumb):
    a_max = np.amax(thumb)
    a_min = np.amin(thumb)
    return np.uint8(255 * (np.float32(thumb) - a_min) / (a_max - a_min))


def recording(dt, is_recording):
    if is_recording:
        logging.info("Recording started at %s", datetime.fromtimestamp(dt))
    else:
        logging.info("Recording ended at %s", datetime.fromtimestamp(dt))


def track_filtered(clip_id, track_id):
    logging.info("Clip %s has filtered track %s", clip_id, track_id)


def tracking_reprocessed(
    clip_id,
    track_id,
    prediction,
    what,
    confidence,
    region,
    frame,
    mass,
    blank,
    tracking,
    last_prediction_frame,
    model_id,
    rec_end,
):
    rec_end = datetime.fromtimestamp(rec_end)
    logging.info(
        "Received post processing event for recording ended at %s clip %s track %s prediction of %s with %s%% confidence at region %s",
        rec_end,
        clip_id,
        track_id,
        what,
        confidence,
        region,
    )

    bus = dbus.SystemBus()
    dbus_object = bus.get_object(DBUS_NAME, DBUS_PATH)
    try:
        thumb, track_id, region = dbus_object.GetThumbnail(clip_id, track_id)
        thumb = np.uint16(thumb)
        thumb = normalize(thumb)
        cv2.imwrite(f"./thumbnails/{rec_end}-{clip_id}-{track_id}.png", thumb)
        logging.info(
            "Saved thumb to %s ", f"./thumbnails/{rec_end}-{clip_id}-{track_id}.png"
        )
    except:
        logging.info(
            "Could not get thumbnail for %s %s", clip_id, track_id, exc_info=True
        )


# helper class to run dbus in background
class TrackingService:
    def __init__(self):
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        self.loop = GLib.MainLoop()
        self.t = threading.Thread(
            target=self.run_server,
        )
        self.t.start()

    def quit(self):
        self.loop.quit()

    def run_server(self):
        dbus_object = None
        try:
            bus = dbus.SystemBus()
            dbus_object = bus.get_object(DBUS_NAME, DBUS_PATH)
        except dbus.exceptions.DBusException as e:
            print("Failed to initialize D-Bus object: '%s'" % str(e))
            sys.exit(2)
        global model_labels
        model_labels = dbus_object.ClassificationLabels()

        bus.add_signal_receiver(
            tracking,
            dbus_interface=DBUS_NAME,
            signal_name="Tracking",
        )
        bus.add_signal_receiver(
            recording,
            dbus_interface=DBUS_NAME,
            signal_name="Recording",
        )

        bus.add_signal_receiver(
            tracking_reprocessed,
            dbus_interface=DBUS_NAME,
            signal_name="TrackingReprocessed",
        )
        bus.add_signal_receiver(
            track_filtered,
            dbus_interface=DBUS_NAME,
            signal_name="TrackFiltered",
        )
        self.loop.run()


def main():
    init_logging()
    thumb_dir = Path("./thumbnails")
    thumb_dir.mkdir(exist_ok=True)
    service = TrackingService()

    # just to keep program alive
    while service.t.is_alive():
        time.sleep(10)


if __name__ == "__main__":
    main()



from flask import Flask
from flask import request
from flask import Response
import numpy as np
import sys
from ml_tools.interpreter import get_interpreter
from config.config import Config
import logging
from ml_tools.logs import init_logging
from config.thermalconfig import ThermalConfig
from waitress import serve

app = Flask(__name__)
# app.config["MAX_CONTENT_LENGTH"] = 407200
interpreter = None
input_shape = None


@app.route("/predict", methods=["POST"])
def main():
    data = request.data
    input_data = np.frombuffer(data, dtype=np.float32)
    input_data = input_data.reshape(input_shape)
    if interpreter.TYPE == "TFLite":
        predictions = []
        for input in input_data:
            prediction = interpreter.predict([input])
            predictions.append(prediction[0])
        predictions = np.array(predictions)
    else:
        predictions = interpreter.predict(input_data)
    response = Response(predictions.tostring(), mimetype="application/octet-stream")
    return response


def startup_classifier():
    # classifies an empty frame to force loading of the model into memory
    num_inputs, in_shape = interpreter.shape()
    if num_inputs > 1:
        zero_input = []
        for shape in in_shape:
            zero_input.append(np.zeros((1, *shape[1:]), np.float32))
    else:
        zero_input = np.zeros((1, *in_shape[1:]), np.float32)
    interpreter.predict(zero_input)


def get_model():

    thermal_config = ThermalConfig.load_from_file()

    if not thermal_config.motion.run_classifier:
        logging.info("Classifier isn't configured to run in config")
        return None

    config = Config.load_from_file()
    network_model = [
        model for model in config.classify.models if model.run_over_network
    ]
    if len(network_model) == 0:
        logging.info("No network model configured in classifier.yaml")
        return None
    if len(network_model) > 1:
        logging.info("Got multiple network models using first")
    return network_model[0]


def main():
    init_logging()
    global interpreter
    global input_shape
    network_model = get_model()
    if network_model is None:
        sys.exit(0)
    interpreter = get_interpreter(network_model)
    num_inputs, input_shape = interpreter.shape()
    if num_inputs > 1:
        logging.error("Not support multiple input models")
        sys.exit(0)
    input_shape = (-1, *input_shape[1:])
    startup_classifier()
    # make sure only 1 thread at a time as classifier is not thread safe
    serve(app, port=network_model.port, threads=1)


if __name__ == "__main__":
    main()



def pi_classify():
    from piclassifier.piclassify import main

    main()


def serve_model():
    from piclassifier.servemodel import main

    main()


def postprocess_watcher():
    from piclassifier.postprocess import main

    main()


def dbus_listener():
    from piclassifier.dbuslistener import main

    main()
